import re
import urllib.error
import urllib.request
import urllib.parse
from PIL import Image
import imghdr
import posixpath

'''
Forked version of Python API to download images from Bing.
Original Author: Guru Prasad (g.gaurav541@gmail.com)
Modified By: Loglux
Updated save_image function: Now it includes exception handling for HTTPError and URLError.
Added features: Resizing, file types, size filters, and label_filename
'''


class Bing:
    def __init__(self, label, limit, output_dir, adult, timeout, filter='', size='', verbose=True,
                 label_filename=False):
        self.download_count = 0
        self.label = label
        self.output_dir = output_dir
        self.adult = adult
        self.filter = filter
        self.verbose = verbose
        self.seen = set()
        self.size = size
        self.use_label_as_filename = label_filename
        assert type(limit) == int, "limit must be integer"
        self.limit = limit
        assert type(timeout) == int, "timeout must be integer"
        self.timeout = timeout
        self.page_counter = 0
        self.headers = {'User-Agent': 'Mozilla/5.0 (X11; Linux x86_64) '
                                      'AppleWebKit/537.11 (KHTML, like Gecko) '
                                      'Chrome/23.0.1271.64 Safari/537.11',
                        'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8',
                        'Accept-Charset': 'ISO-8859-1,utf-8;q=0.7,*;q=0.3',
                        'Accept-Encoding': 'none',
                        'Accept-Language': 'en-US,en;q=0.8',
                        'Connection': 'keep-alive'}

    def get_filter(self, shorthand):
        if shorthand == "line" or shorthand == "linedrawing":
            return "+filterui:photo-linedrawing"
        elif shorthand == "photo":
            return "+filterui:photo-photo"
        elif shorthand == "clipart":
            return "+filterui:photo-clipart"
        elif shorthand == "gif" or shorthand == "animatedgif":
            return "+filterui:photo-animatedgif"
        elif shorthand == "transparent":
            return "+filterui:photo-transparent"
        else:
            return ""

    def save_image(self, link, file_path, resize_dim=None):
        try:
            request = urllib.request.Request(link, None, self.headers)
            response = urllib.request.urlopen(request, timeout=self.timeout)
        except urllib.error.HTTPError as e:
            # The server couldn't fulfill the request (i.e., a 404 or 500 error)
            print('HTTPError: {}, link: {}'.format(e.code, link))
            raise
        except urllib.error.URLError as e:
            # Failed to reach the server (i.e., the server is down or the URL is incorrect)
            print('URLError: {}, link: {}'.format(e.reason, link))
            raise
        else:
            # If there were no errors, then proceed to saving the image
            image = response.read()
            if not imghdr.what(None, image):
                print('[Error]Invalid image, not saving {}\n'.format(link))
                raise ValueError('Invalid image, not saving {}\n'.format(link))
            with open(str(file_path), 'wb') as f:
                f.write(image)
            if resize_dim:
                img = Image.open(file_path)
                img = img.resize(resize_dim, Image.LANCZOS)
                img.save(file_path)

    def download_image(self, link, resize_dim=None, file_types=None):
        if file_types:
            current_file_type = link.split('.')[-1].lower()
            if current_file_type not in file_types:
                return  # skip this image, as it doesn't match the desired file types
        self.download_count += 1
        # Get the image link
        try:
            path = urllib.parse.urlsplit(link).path
            filename = posixpath.basename(path).split('?')[0]
            file_type = filename.split(".")[-1]
            if file_type.lower() not in ["jpe", "jpeg", "jfif", "exif", "tiff", "gif", "bmp", "png", "webp", "jpg"]:
                file_type = "jpg"

            if self.verbose:
                # Download the image
                print("[%] Downloading Image #{} from {}".format(self.download_count, link))

            # determine file name based on the flag
            if self.use_label_as_filename:
                file_name = f"{self.label.replace(' ', '_')}_{str(self.download_count).zfill(4)}.{file_type}"
            else:
                file_name = f"Image_{str(self.download_count).zfill(4)}.{file_type}"

            self.save_image(
                link,
                self.output_dir.joinpath(file_name),
                resize_dim
            )

            if self.verbose:
                print("[%] File Downloaded !\n")

        except Exception as e:
            self.download_count -= 1
            print("[!] Issue getting: {}\n[!] Error:: {}".format(link, e))

    def run(self, file_types=None, resize_dim=None):
        while self.download_count < self.limit:
            if self.verbose:
                print('\n\n[!!]Indexing page: {}\n'.format(self.page_counter + 1))
            # Parse the page source and download pics
            request_url = 'https://www.bing.com/images/async?q=' + urllib.parse.quote_plus(self.label) \
                          + '&first=' + str(self.page_counter) + '&count=' + str(self.limit) \
                          + '&adlt=' + self.adult + '&qft=' + (
                              '' if self.filter is None else self.get_filter(self.filter))
            if self.size:
                request_url += f'+filterui:imagesize-{self.size}'
            request = urllib.request.Request(request_url, None, headers=self.headers)
            response = urllib.request.urlopen(request)
            html = response.read().decode('utf8')
            if html == "":
                print("[%] No more images are available")
                break
            links = re.findall('murl&quot;:&quot;(.*?)&quot;', html)
            if self.verbose:
                print("[%] Indexed {} Images on Page {}.".format(len(links), self.page_counter + 1))
                print("\n===============================================\n")

            for link in links:
                if self.download_count < self.limit and link not in self.seen:
                    self.seen.add(link)
                    self.download_image(link, resize_dim, file_types)

            self.page_counter += 1
        print("\n\n[%] Done. Downloaded {} images.".format(self.download_count))
